use std::sync::Arc;
use tokio::sync::{Mutex, MutexGuard};
use tokio::time::{timeout, Duration};

/// Provides a 30 second timeout for lock attempts
const LOCK_TIMEOUT: u64 = 30;

#[derive(Clone)]
pub struct Shared<T> {
    pub data: Arc<Mutex<T>>,
}

impl<T> Shared<T> {
    /// Creates a new Shared of data T
    ///
    /// > Want to see a complete example, look at doc_example in the examples directory
    /// 
    /// Example:
    /// ```
    /// use shared::Shared;
    ///
    /// #[derive(Clone)]
    /// struct Person {
    ///     pub name: String,
    ///     pub age: u16
    /// }
    ///
    /// #[tokio::main]
    /// async fn main() {
    ///     let shared_data = Shared::new(Person{
    ///         name: "Test Dummy".to_string(),
    ///         age: 30
    ///     });
    ///
    ///     // See Shared::lock for using it
    /// }
    /// ```
    pub fn new(data: T) -> Self {
        Shared {
            data: Arc::new(Mutex::new(data)),
        }
    }

    /// Acquire lock
    /// 
    /// > Want to see a complete example, look at doc_example in the examples directory
    ///
    /// Using:
    /// ```
    /// // See Shared::new for creating a Shared<Person>
    /// 
    /// { // Direct access
    ///     let person = shared_data.lock().await;
    ///     println!("Person{{name: {}, age: {}}}", person.name, person.age);
    /// }
    /// 
    /// // Using it as a parameter
    /// async fn get_older(shared: Shared<Person>) {
    ///     let mut person = shared.lock().await;
    ///     person.age += 1;
    /// }
    /// 
    /// // Calling
    /// get_older(shared_data.clone()).await;
    /// ```
    pub async fn lock(&self) -> MutexGuard<'_, T> {
        let result = timeout(Duration::from_secs(LOCK_TIMEOUT), self.data.lock()).await;
        if let Ok(guard) = result {
            return guard;
        }
        panic!(
            "Failed to acquire lock in {} seconds.  Deadlock?",
            LOCK_TIMEOUT
        );

        // self.data.lock().await
    }
}
